        <div class="outer">
            <br><var>command</var><br>
            <div class="codeblock">
		<pre><code>{{cmd}}</code></pre>
            </div>

            <var>stdout</var><br>
            <div class="output">

		<pre><samp>{{stdout}}</samp></pre>
            </div>

            <var>stderr</var><br>
            <div class="error">

		<pre><samp>{{stderr}}</samp></pre>
            </div>
            
            
            <table class="statustable" >
                <tr>
                    <th>Program</th>
                    <th>Return code</th>
                    <th>Start</th>
                    <th>Runtime</th>
                    <th>Version</th>
                </tr>
                <tr>
		    <td>{{name}}</td>
		    <td style="background-color:{{statuscolor}};">{{exitcode}}</td>
		    <td>{{starttime}}</td>
		    <td>{{runtime}}</td>
                    <td>{{version}}</td>
                </tr>
                
            </table>
	
	    <br><br>
	<hr>


        </div>
