/*
To change this license header, choose License Headers in Project Properties.
To change this template file, choose Tools | Templates
and open the template in the editor.
*/
/* 
    Created on : Dec 11, 2019, 8:11:42 PM
    Author     : usingh
*/
pre {
    overflow-x: auto;
    white-space: pre-wrap;
    white-space: -moz-pre-wrap;
    white-space: -pre-wrap;
    white-space: -o-pre-wrap;
    word-wrap: break-word;
}
.outer{
    background:appworkspace;
    margin-left: 20px;
    margin-right: 20px;
}
.codeblock{
    background: rgb(245,245,245);
    margin: auto;
}
.output{
    background: rgb(220,235,233);
}
.error{
    color: red;
    background: linen;
}
table, th, td {
    border: 1px solid black;
}
.statustable{
    margin: auto;
    border: 1px solid black;
    border-collapse: collapse;
    width:100%;
} 
.programinfo{
    border-collapse: collapse;
}
.sysInfotable{
    border-collapse: collapse;
}
.sysmodules,.syspath{
    border-collapse: collapse;
}
.envtabs{
    margin-left: 20px;
    margin-right: 20px;
}